/*
	I am something like SHA1 - equally secure and running in parallel.

	To run me you need machine with SSE, 32bit int, gcc, mhash and good
	luck. Setting THREADS according to the number of your CPUs might also
	help. If you want to know more about me, see README.

	Anicka <anicka@anicka.net> wrote me in February 2009
	and wishes to forget that she ever messed with me.

	Use me according to terms of GPLv2.
*/

#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <mhash.h>
#include <unistd.h>
#include <fcntl.h>
#include <emmintrin.h>

#include "common.h"

#define HASHSIZE 20 	/* SHA is 160 bits long */
#define THREADS 4 	/* set me */
#define MAX_FIFO (4*THREADS)

#define BUFSIZE (4 * 4 * 16) /* length of one sha block */
#define WORKSIZE (4 * 4 * 80) /* sha workspace */
#define STEP 65536 /* size of the read buffer */

/* SHA magic numbers */
#define H0 0x67452301
#define H1 0xEFCDAB89
#define H2 0x98BADCFE
#define H3 0x10325476
#define H4 0xC3D2E1F0

#define K1 0x5A827999
#define K2 0x6ED9EBA1
#define K3 0x8F1BBCDC
#define K4 0xCA62C1D6

struct hash {
	unsigned char h[HASHSIZE];
};

struct block {
	char data[STEP];
	int seqn;
	struct hash h;
	int dsize;
};

struct fifo {
	struct block* bl[MAX_FIFO];
	sem_t sem;
	pthread_mutex_t mut;
	int end;
	int start; 
};

__thread char __attribute__ ((aligned (16))) buf[WORKSIZE];
__thread unsigned int __attribute__ ((aligned (16))) h0[] = {H0, H0, H0, H0};
__thread unsigned int __attribute__ ((aligned (16))) h1[] = {H1, H1, H1, H1};
__thread unsigned int __attribute__ ((aligned (16))) h2[] = {H2, H2, H2, H2};
__thread unsigned int __attribute__ ((aligned (16))) h3[] = {H3, H3, H3, H3};
__thread unsigned int __attribute__ ((aligned (16))) h4[] = {H4, H4, H4, H4};

struct fifo input, output;
pthread_t* threads;
int fd;
struct hash* hashes;
int hashes_size;
unsigned char result[HASHSIZE];
int seq; /* at the end it is number of blocks + 2 */

/* sha rotation */
static inline __m128i
circ(__m128i a, int b)
{
	__m128i f, g;

	f = _mm_slli_epi32(a, b);
	g = _mm_srli_epi32(a, 32 - b);
	return _mm_or_si128(f, g);
}

static void
process_block(void)
{
	int i;
	__m128i k, t, u, v, w, x, y, z, temp, f, g;
	__m128i a, b, c, d, e;
	__m128i* pb = (__m128i*) buf;

	a = _mm_load_si128((__m128i*) h0);
	b = _mm_load_si128((__m128i*) h1);
	c = _mm_load_si128((__m128i*) h2);
	d = _mm_load_si128((__m128i*) h3);
	e = _mm_load_si128((__m128i*) h4);

	for (i = 16; i <= 79; i++) {
		y = _mm_load_si128(pb + (i - 3));
		z = _mm_load_si128(pb + (i - 8));
		u = _mm_load_si128(pb + (i - 14));
		v = _mm_load_si128(pb + (i - 16));

		x = _mm_xor_si128(y, z); 
		x = _mm_xor_si128(x, u); 
		x = _mm_xor_si128(x, v);

		x = circ(x, 1);
		
		_mm_store_si128(pb + i, x);
	}
	for (i = 0; i < 20; i++) {
		w = _mm_load_si128(pb + i);
		k = _mm_set1_epi32(K1);

		temp = _mm_add_epi32(e, w);
		temp = _mm_add_epi32(temp, k);
		f = _mm_and_si128(b, c); 
		g = _mm_andnot_si128(b, d); 
		f = _mm_or_si128(f, g);
		temp = _mm_add_epi32(temp, f);
		g = circ(a, 5);
		temp = _mm_add_epi32(temp, g);
		
		e = d;
		d = c;
		c = circ(b, 30);
		b = a;
		a = temp;
	}
	for (i = 20; i < 40; i++) {
		w = _mm_load_si128(pb + i);
		k = _mm_set1_epi32(K2);

		temp = _mm_add_epi32(e, w);
		temp = _mm_add_epi32(temp, k);
		f = _mm_xor_si128(b, c); 
		f = _mm_xor_si128(f, d);
		temp = _mm_add_epi32(temp, f);
		g = circ(a, 5);
		temp = _mm_add_epi32(temp, g);
		
		e = d;
		d = c;
		c = circ(b, 30);
		b = a;
		a = temp;
	}
	for (i = 40; i < 60; i++) {
		w = _mm_load_si128(pb + i);
		k = _mm_set1_epi32(K3);

		temp = _mm_add_epi32(e, w);
		temp = _mm_add_epi32(temp, k);
		f = _mm_and_si128(b, c); 
		g = _mm_and_si128(b, d); 
		f = _mm_or_si128(f, g);
		g = _mm_and_si128(c, d); 
		f = _mm_or_si128(f, g);
		temp = _mm_add_epi32(temp, f);
		g = circ(a, 5);
		temp = _mm_add_epi32(temp, g);
		
		e = d;
		d = c;
		c = circ(b, 30);
		b = a;
		a = temp;
	}
	for (i = 60; i < 80; i++) {
		w = _mm_load_si128(pb + i);
		k = _mm_set1_epi32(K4);

		temp = _mm_add_epi32(e, w);
		temp = _mm_add_epi32(temp, k);
		f = _mm_xor_si128(b, c); 
		f = _mm_xor_si128(f, d);
		temp = _mm_add_epi32(temp, f);
		g = circ(a, 5);
		temp = _mm_add_epi32(temp, g);
		
		e = d;
		d = c;
		c = circ(b, 30);
		b = a;
		a = temp;
	}
	t = _mm_load_si128((__m128i*) h0);
	t = _mm_add_epi32(t, a);
	_mm_store_si128((__m128i*) h0, t);

	t = _mm_load_si128((__m128i*) h1);
	t = _mm_add_epi32(t, b);
	_mm_store_si128((__m128i*) h1, t);

	t = _mm_load_si128((__m128i*) h2);
	t = _mm_add_epi32(t, c);
	_mm_store_si128((__m128i*) h2, t);

	t = _mm_load_si128((__m128i*) h3);
	t = _mm_add_epi32(t, d);
	_mm_store_si128((__m128i*) h3, t);

	t = _mm_load_si128((__m128i*) h4);
	t = _mm_add_epi32(t, e);
	_mm_store_si128((__m128i*) h4, t);
}

static struct block*
new_block(int s)
{
	struct block* pb;

	pb = xmemalign(16, sizeof(struct block));
	pb->seqn = s;		
	pb->dsize = read(fd, &pb->data, STEP);
	if (pb->dsize < 0)
		die("read: %m\n");
	return pb;
}

static void
push(struct block* bl, struct fifo* target)
{
	pthread_mutex_lock(&target->mut);
	target->bl[target->end] = bl;
	target->end = (target->end + 1) % MAX_FIFO;
	pthread_mutex_unlock(&target->mut);
	sem_post(&target->sem);
}

static struct block*
pop(struct fifo* from)
{
	struct block* pb;
	sem_wait(&from->sem);
	
	pthread_mutex_lock(&from->mut);
	pb = from->bl[from->start];
        from->start = (from->start + 1) % MAX_FIFO;
        pthread_mutex_unlock(&from->mut);

	return pb;
}

static void
init_fifo(struct fifo* f)
{
	sem_init(&f->sem, 0, 0);
	pthread_mutex_init(&f->mut, NULL);
	f->start = 0;
	f->end = 0;
}

static void
init(void)
{
	threads = xmalloc(THREADS * sizeof(pthread_t));
	hashes_size = STEP;
	hashes = xmalloc(hashes_size * sizeof(struct hash));
	init_fifo(&input);
	init_fifo(&output);
}

static void
store(struct block* pb)
{
	if (pb->seqn >= hashes_size) {
		hashes_size = 2 * (pb->seqn + 1);
		hashes = xrealloc(hashes, hashes_size * sizeof(struct hash));
	}
	hashes[pb->seqn] = pb->h;		
}

static void
finish(void)
{
	sem_destroy(&input.sem);
	sem_destroy(&output.sem);
	pthread_mutex_destroy(&input.mut);
	pthread_mutex_destroy(&output.mut);
}

static void*
hashme(void* id)
{
	MHASH h;
	struct block* pb;
	char* p;
	int s;
	int i;

	for (;;) {
		_mm_store_si128((__m128i*)h0, _mm_set1_epi32(H0));
		_mm_store_si128((__m128i*)h1, _mm_set1_epi32(H1));
		_mm_store_si128((__m128i*)h2, _mm_set1_epi32(H2));
		_mm_store_si128((__m128i*)h3, _mm_set1_epi32(H3));
		_mm_store_si128((__m128i*)h4, _mm_set1_epi32(H4));
		
		pb = pop(&input);
		if(pb->dsize <= 0) {
			push(pb, &output);
			break;	
		}		
		s = pb->dsize;
		p = pb->data;
		while (s > 0) {
			if (s < BUFSIZE)
				memset(p + s, 0, BUFSIZE - s);
			for (i = 0; i < BUFSIZE; i += 16){	
				_mm_store_si128((__m128i*)(buf + i),
					_mm_load_si128((__m128i*)(p + i)));
			}
			process_block();
			s -= BUFSIZE;
			p += BUFSIZE;
		}
		h = mhash_init(MHASH_SHA1);
		if (h == MHASH_FAILED)
			die("mhash_init failed");
		mhash(h, h0, 4 * sizeof(unsigned int));
		mhash(h, h1, 4 * sizeof(unsigned int));
		mhash(h, h2, 4 * sizeof(unsigned int));
		mhash(h, h3, 4 * sizeof(unsigned int));
		mhash(h, h4, 4 * sizeof(unsigned int));
		mhash(h, &pb->dsize, sizeof(int));
		mhash_deinit(h, pb->h.h);
		push(pb, &output);
	}
	pthread_exit(NULL);
}

static void
hash_result(void)
{
	MHASH h;
	int i;

	h = mhash_init(MHASH_SHA1);
	if (h == MHASH_FAILED)
		die("mhash_init failed");
	for (i = 0; i <= seq - 2; i++)
		mhash(h, &hashes[i].h, HASHSIZE);
	mhash_deinit(h, result);
}

int
main(int argc, char** argv)
{
	int i, err, flag = 0;
	int total_req = 0;
	struct block* pb;

	if (argc > 1) {
		fd = open(argv[1], O_RDONLY);
		if (fd < 0)
			die("open: %m");	
	}

	init();
	for (i = 0; i < THREADS; i++) {
		err = pthread_create(&threads[i], NULL, 
			hashme, (void*)(intptr_t)i);
                if (err)
                        die("pthread_create: %m");
	}

	for (;;) {
		while(total_req < MAX_FIFO && !flag) {
			pb = new_block(seq++);
			if (pb->dsize <=0) {
				flag = 1;
				free(pb);
			} else {
				push(pb, &input);
				total_req++;
			}
		}
		if (flag && !total_req)
			break;
		pb = pop(&output);
		total_req--;

		store(pb);
		free(pb);
	}
	for (i = 0; i < THREADS; i++) {
		pb = xmalloc(sizeof(struct block));
		pb->dsize = 0;
		push(pb, &input);
	}
	for (i = 0; i < THREADS; i++)
		free(pop(&output));

	hash_result();
	for (i = 0; i < mhash_get_block_size(MHASH_SHA1); i++)
                       printf("%.2x", result[i]);
	putchar('\n');

	finish();
	pthread_exit(NULL);
}
